@@ -136,7 +136,7 @@ void
 Controller::initializePoolOptions(Client *client, Request *req, RequestAnalysis &analysis) {
 	boost::shared_ptr<Options> *options;
 
-	if (req->config->singleAppMode) {
+	if (mainConfig.singleAppMode) {
 		P_ASSERT_EQ(poolOptionsCache.size(), 1);
 		poolOptionsCache.lookupRandom(NULL, &options);
 		req->options = **options;
@@ -184,8 +184,8 @@ Controller::fillPoolOptionsFromConfigCaches(Options &options,
 	options.ruby = requestConfig->defaultRuby;
 	options.nodejs = requestConfig->defaultNodejs;
 	options.python = requestConfig->defaultPython;
-	options.meteorAppSettings = requestConfig->meteorAppSettings;
-	options.fileDescriptorUlimit = requestConfig->fileDescriptorUlimit;
+	options.meteorAppSettings = requestConfig->defaultMeteorAppSettings;
+	options.fileDescriptorUlimit = requestConfig->defaultAppFileDescriptorUlimit;
 
 	options.logLevel = int(LoggingKit::getLevel());
 	options.integrationMode = psg_pstrdup(pool, mainConfig.integrationMode);
@@ -195,15 +195,16 @@ Controller::fillPoolOptionsFromConfigCaches(Options &options,
 	options.userSwitching = mainConfig.userSwitching;
 	options.defaultUser = requestConfig->defaultUser;
 	options.defaultGroup = requestConfig->defaultGroup;
-	options.minProcesses = requestConfig->minInstances;
-	options.maxPreloaderIdleTime = requestConfig->maxPreloaderIdleTime;
-	options.maxRequestQueueSize = requestConfig->maxRequestQueueSize;
-	options.abortWebsocketsOnProcessShutdown = requestConfig->abortWebsocketsOnProcessShutdown;
-	options.forceMaxConcurrentRequestsPerProcess = requestConfig->forceMaxConcurrentRequestsPerProcess;
-	options.spawnMethod = requestConfig->spawnMethod;
-	options.loadShellEnvvars = requestConfig->loadShellEnvvars;
+	options.minProcesses = requestConfig->defaultMinInstances;
+	options.maxPreloaderIdleTime = requestConfig->defaultMaxPreloaderIdleTime;
+	options.maxRequestQueueSize = requestConfig->defaultMaxRequestQueueSize;
+	options.abortWebsocketsOnProcessShutdown = requestConfig->defaultAbortWebsocketsOnProcessShutdown;
+	options.forceMaxConcurrentRequestsPerProcess = requestConfig->defaultForceMaxConcurrentRequestsPerProcess;
+	options.environment = requestConfig->defaultEnvironment;
+	options.spawnMethod = requestConfig->defaultSpawnMethod;
+	options.loadShellEnvvars = requestConfig->defaultLoadShellEnvvars;
 	options.statThrottleRate = mainConfig.statThrottleRate;
-	options.maxRequests = requestConfig->maxRequests;
+	options.maxRequests = requestConfig->defaultMaxRequests;
 
 	/******************************/
 }
@@ -473,15 +474,13 @@ Controller::onRequestBegin(Client *client, Request *req) {
 		// and localize them as much as possible, for better CPU caching.
 		RequestAnalysis analysis;
 		analysis.flags = req->secureHeaders.lookup(FLAGS);
-		analysis.appGroupNameCell = req->config->singleAppMode
+		analysis.appGroupNameCell = mainConfig.singleAppMode
 			? NULL
 			: req->secureHeaders.lookupCell(PASSENGER_APP_GROUP_NAME);
 		analysis.unionStationSupport = unionStationContext != NULL
 			&& getBoolOption(req, UNION_STATION_SUPPORT, false);
 		req->stickySession = getBoolOption(req, PASSENGER_STICKY_SESSIONS,
-			mainConfig.stickySessions);
-		req->showVersionInHeader = getBoolOption(req, PASSENGER_SHOW_VERSION_IN_HEADER,
-			req->config->showVersionInHeader);
+			mainConfig.defaultStickySessions);
 		req->host = req->headers.lookup(HTTP_HOST);
 
 		/***************/
